#!/usr/bin/env python3
__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import pickle
from argparse import ArgumentParser, Namespace
from pathlib import Path
from sys import argv
from typing import List, Dict, Type, Any

from media_tools.util import (
    DuplicateStringRemover, Commands, JunkRemover, NumberingFixer, SymlinkFixer, UndoCommands
)

UNDO_DATABASE_FILE = '~/.music-rename-undo.pickle'


class FilenameCleaner:

    def __init__(
            self, basedir: str, verbose: bool, force: bool, undo_db: str = UNDO_DATABASE_FILE
    ) -> None:
        self._base_directory = Path(basedir)
        self._verbose = verbose
        self._force = force
        self._undo_db = Path(undo_db).expanduser()
        self._undo_info = self._load_undo_info(self._undo_db)

    def __del__(self) -> None:
        with open(self._undo_db, 'wb') as db_file:
            return pickle.dump(self._undo_info, db_file)

    def clean_filenames(self, min_length: int = 0, recurse: bool = False) -> None:
        self._execute_fix_commands(
            self._get_fix_commands(DuplicateStringRemover, min_length=min_length, recurse=recurse)
        )

    def clean_numbering(self) -> None:
        self._execute_fix_commands(self._get_fix_commands(NumberingFixer))

    def clean_junk(self) -> None:
        self._execute_fix_commands(self._get_fix_commands(JunkRemover))

    def undo(self) -> None:
        self._execute_fix_commands(self._get_fix_commands(UndoCommands, undo_info=self._undo_info))

    def fix_symlinks(self) -> None:
        self._execute_fix_commands(self._get_fix_commands(SymlinkFixer, undo_info=self._undo_info))

    def _get_fix_commands(self, cleaner_class: Type, **kwargs: Any) -> Commands:
        cleaner = cleaner_class(self._base_directory, self._verbose)
        return cleaner.fix_commands(**kwargs)

    def _execute_fix_commands(self, fix_commands: Commands) -> None:
        for fix_command in [f for f in fix_commands if f[0] is not None]:
            self._undo_info[fix_command.source] = fix_command.destination
            fix_command.execute(self._verbose, self._force)
        if self._verbose:
            print(f"{len(fix_commands)} fixed")

    @staticmethod
    def _load_undo_info(undo_db: Path) -> Dict[str, str]:
        try:
            with open(undo_db, 'rb') as db_file:
                return pickle.load(db_file)
        except (FileNotFoundError, EOFError):
            return {}


def parse_commandline(args: List[str]) -> Namespace:
    parser = ArgumentParser(
        description="Perform several operations to remove junk from music file names"
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true'
    )
    parser.add_argument('-f', '--force', action='store_true', help='Force rename files')

    subparsers = parser.add_subparsers(help='Available subcommands:', dest='subparser')
    # workaround for python issue https://bugs.python.org/issue26510
    subparsers.required = True  # type: ignore

    parser_filenames = subparsers.add_parser(
        'clean-filenames', help='Remove longest common substring from music files in this dir'
    )
    parser_filenames.add_argument(
        '-r', '--recurse', action='store_true', help='Scan subdirectories recursively'
    )
    parser_filenames.add_argument(
        '--min-length', type=int, default=5, help='Minimum length for common substring'
    )

    subparsers.add_parser('clean-numbering', help='Let music files start with "%d - "')
    subparsers.add_parser('clean-junk', help='Remove junk from filenames')
    subparsers.add_parser('undo', help='Undo a change on the target directory')
    subparsers.add_parser('fix-symlinks', help='Fix symlinks broken by earlier cleaning steps')

    parser.add_argument('target', type=str)
    return parser.parse_args(args)


def main() -> None:
    opts = parse_commandline(argv[1:])
    cleaner = FilenameCleaner(opts.target, verbose=opts.verbose, force=opts.force)
    if opts.subparser == 'undo':
        cleaner.undo()
    elif opts.subparser == 'clean-filenames':
        cleaner.clean_filenames(min_length=opts.min_length, recurse=opts.recurse)
    elif opts.subparser == 'clean-numbering':
        cleaner.clean_numbering()
    elif opts.subparser == 'clean-junk':
        cleaner.clean_junk()
    elif opts.subparser == 'fix-symlinks':
        cleaner.fix_symlinks()


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import List


class BaseFilenameCleaner:

    MUSIC_EXTENSIONS = ('mp3', 'flac', 'ogg', 'm4a')

    def __init__(self, basedir: Path, verbose: bool) -> None:
        self._base_directory = basedir
        self._verbose = verbose

    def get_music_files(self) -> List[Path]:
        return sorted(
            [
                f for f in Path(self._base_directory).iterdir()
                if f.is_file() and self.is_music_file(f)
            ]
        )

    @staticmethod
    def is_music_file(filename: Path) -> bool:
        return any(
            filename.name.upper().endswith(e.upper()) for e in BaseFilenameCleaner.MUSIC_EXTENSIONS
        )

    @staticmethod
    def filename_base(file: Path) -> str:
        return file.stem



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import sys
from argparse import ArgumentParser, Namespace
from datetime import timedelta
from functools import lru_cache
from pathlib import Path
from typing import List, Dict

import audioread
import magic
from audioread.exceptions import NoBackendError
from mutagen import File, MutagenError

AUDIO_EXTENSIONS = ('mp3', 'ogg', 'flac', 'wav', 'm4a', 'aif', 'aiff')


def parse_commandline(args: List[str]) -> Namespace:
    parser = ArgumentParser(
        description="Print length of audio files in directories"
    )
    parser.add_argument(
        'directory', type=str, help='Directories to print'
    )
    return parser.parse_args(args)


class LengthStore:

    def __init__(self, root: Path):
        self.root = root
        self.lengths: Dict[Path, timedelta] = {}
        self.errors: Dict[Path, Exception] = {}

    def scan(self):
        self.lengths[self.root] = self.dir_length(self.root)

    @lru_cache(maxsize=100000)
    def dir_length(self, folder: Path) -> timedelta:
        length = timedelta(seconds=0)
        for thing in folder.iterdir():
            if is_audio(thing):
                length += timedelta(seconds=self.get_duration(thing))
            elif thing.is_dir():
                duration = self.dir_length(thing)
                if duration:
                    self.lengths[thing] = duration
                    length += duration
        return length

    def print(self) -> None:
        max_duration = str(self.lengths[self.root]).split('.')[0]
        for path, duration in sorted(self.lengths.items(), key=lambda item: item[1]):
            self.print_line(
                path if path == self.root else path.relative_to(self.root),
                str(duration).split('.')[0].rjust(len(max_duration))
            )

    @staticmethod
    def print_line(path: Path, duration: str) -> None:
        print(f"{duration}   {path}")

    def get_duration(self, track: Path) -> float:
        try:
            audio = File(track)
            if audio:
                return audio.info.length
        except MutagenError:
            pass
        try:
            with audioread.audio_open(track) as file:
                return file.duration
        except (NoBackendError, EOFError, UnicodeEncodeError) as error:
            self.errors[track] = error
            return 0


def is_audio(track: Path):
    if not track.is_file():
        return False
    if track.suffix[1:].lower() in AUDIO_EXTENSIONS:
        return True
    file_magic = magic.from_file(str(track), mime=True)
    return file_magic.startswith('audio/')


def main() -> None:
    args: List[str] = sys.argv[1:]
    opts = parse_commandline(args)
    lengths = LengthStore(Path(opts.directory))
    lengths.scan()
    lengths.print()
    if lengths.errors:
        print(len(lengths.errors), 'errors')


if __name__ == '__main__':
    main()



from pathlib import Path
from shutil import move
from typing import Any, List, NamedTuple

from .util import print_utf8_safe


class Command(NamedTuple):
    src: Path = None
    dest: Path = None
    info: List[Any] = []

    @property
    def source(self) -> str:
        return str(self.src.resolve())

    @property
    def destination(self) -> str:
        return str(self.dest.resolve())

    def execute(self, verbose: bool, force: bool) -> None:
        if force:
            self.do_execute(verbose)
        self.print(verbose)

    def print(self, verbose: bool) -> None:
        raise NotImplementedError()

    def do_execute(self, verbose: bool) -> None:
        raise NotImplementedError()


class Nothing(Command):
    def do_execute(self, verbose: bool) -> None:
        pass

    def print(self, verbose: bool) -> None:
        pass


class Move(Command):
    def do_execute(self, verbose: bool) -> None:
        try:
            move(self.source, self.destination)
        except FileNotFoundError:
            if verbose:
                print_utf8_safe('FAIL:', self.source, '->', self.destination)

    def print(self, verbose: bool) -> None:
        if verbose:
            print(f'mv {self.source} {self.destination} # {self.info}')


class Relink(Command):
    @property
    def new_symlink(self) -> Path:
        return self.src.parent.joinpath(self.dest.name)

    def do_execute(self, verbose: bool) -> None:
        self.src.unlink()
        self.new_symlink.symlink_to(self.dest)

    def print(self, verbose: bool) -> None:
        if verbose:
            print(f'rm -f {self.source}')
            print(f'ln -s {self.destination} {str(self.new_symlink)}')


Commands = List[Command]



from pathlib import Path
from typing import Dict

from .base_filename_cleaner import BaseFilenameCleaner
from .command import Commands, Move


class UndoCommands(BaseFilenameCleaner):

    def fix_commands(self, undo_info: Dict[str, str]) -> Commands:
        path = self._base_directory.resolve()
        fix_commands: Commands = []
        for source, destination in [
                (dest, src) for src, dest in undo_info.items() if dest.startswith(str(path))
        ]:
            if Path(source).is_file():
                fix_commands.append(Move(Path(source), Path(destination)))
            elif self._verbose:
                print('FAIL', source)
        return fix_commands



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from argparse import ArgumentParser, Namespace
from datetime import datetime
from pathlib import Path
from sys import argv
from typing import List

from media_tools.util.buying.track import Track, TrackDB
from media_tools.util.buying.distributor import Beatport, Amazon
from media_tools.util.buying.last_fm import LastFM

DEFAULT_TRACK_DB = 'bought.pickle'


def read_datetime(args: str) -> datetime:
    try:
        return datetime.strptime(args, '%Y-%m-%d')
    except ValueError:
        print('Accepted date format: YYYY-MM-DD')
        raise


def parse_commandline(args: List[str]) -> Namespace:
    parser = ArgumentParser(
        description="..."
    )
    parser.add_argument(
        '-u', '--user', required=True, help='Last.FM username'
    )
    parser.add_argument(
        '-k', '--api-key', required=True, help='Last.FM API key'
    )
    parser.add_argument(
        '-l', '--limit', type=int, default=10, help='Maximum number of tracks to display'
    )
    parser.add_argument(
        '-m', '--min-plays', type=int, default=1, help='Minimum number of plays per track'
    )
    parser.add_argument(
        '--buy-up-to', type=int, default=None,
        help='When specified, repeat until this may tracks have been bought'
    )
    period_group = parser.add_mutually_exclusive_group()
    period_group.add_argument(
        '-p', '--period', choices=('overall', '7day', '1month', '3month', '6month', '12month'),
        default='overall', help='Period from which to choose favorite tracks'
    )
    period_group.add_argument(
        '-f', '--from-date', type=read_datetime,
        help='...'
    )
    parser.add_argument(
        '-t', '--to-date', type=read_datetime, default=datetime.now(),
        help='...'
    )
    parser.add_argument(
        '-d', '--track-db', type=str, default=DEFAULT_TRACK_DB,
        help='Name of the file storing already bought tracks'
    )

    return parser.parse_args(args)


def main() -> None:
    opts = parse_commandline(argv[1:])
    api = LastFM(opts.user, opts.api_key)

    with TrackDB(Path(opts.track_db)) as trackdb:
        Track.setup((Beatport(), Amazon()), trackdb)
        bought = buy_exactly(api, opts)
    print(bought, 'bought')


def buy_exactly(api: LastFM, opts: Namespace, verbose: bool = False) -> int:
    bought = 0
    while bought < (opts.buy_up_to or 1):
        for track in get_tracks_chunk(api, opts):
            if verbose:
                print(track, track.buy_url())
            track.buy()
            if track.buy_url():
                bought += 1
                if opts.buy_up_to and bought >= opts.buy_up_to:
                    break
    return bought


def get_tracks_chunk(api: LastFM, opts: Namespace, page: int = 1) -> List[Track]:
    return api.get_tracks_by_period(
        from_date=opts.from_date, to_date=opts.to_date, limit=opts.limit, page=page,
        min_plays=opts.min_plays
    ) if opts.from_date else api.get_top_tracks(
        period=opts.period, limit=opts.limit, page=page, min_plays=opts.min_plays
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import os
import re
from argparse import ArgumentParser, Namespace
from math import log10
from pathlib import Path
from shutil import copy2, SameFileError
from sys import argv
from time import time
from typing import List

from media_tools.util import AudaciousTools, find_files


def copy_playlist(  # pylint: disable=too-many-arguments
        playlist_id: str, number: int, target: str, verbose: bool = False, renumber: bool = False,
        audacious: AudaciousTools = None
) -> None:
    if not os.path.isdir(target):
        os.mkdir(target)

    if audacious is None:
        audacious = AudaciousTools()

    playlist_id = playlist_id or audacious.get_currently_playing_playlist_id()

    copy_files(audacious.get_files_to_copy(number, playlist_id), target, verbose, renumber)


def strip_leading_numbers(filename: str) -> str:
    return re.sub(r'^\d+\s*[-.]?\s*', '', filename)


def renumber_file(filename: str, number: int, total: int) -> str:
    return "{:0{width}d} - {}".format(
        number, strip_leading_numbers(filename), width=max(int(log10(total)) + 1, 2)
    )


def copy_files(files_to_copy: List[str], target_dir: str, verbose: bool, renumber: bool) -> None:
    for i, file in enumerate(files_to_copy):
        filename = file.split('/')[-1]
        target_filename = renumber_file(filename, i + 1, len(files_to_copy)) if renumber \
            else filename
        if verbose:
            print("{}/{}: {}".format(i + 1, len(files_to_copy), target_filename))
        copy_file(file, os.path.join(target_dir, target_filename))


def copy_file(file: str, target: str) -> None:
    try:
        copy2(file, target)
    except SameFileError as error:
        print(str(error))


def move_files_to_original_places(
        playlist_id: str, music_dir: str = '/home/preuss/Music', verbose: bool = False,
        audacious: AudaciousTools = None
) -> None:

    def find(name: str, path: str) -> str:
        for root, _, files in os.walk(path):
            if name in files:
                return os.path.join(root, name)
        return ''

    if audacious is None:
        audacious = AudaciousTools()

    playlist_id = playlist_id or audacious.get_currently_playing_playlist_id()
    for file in audacious.files_in_playlist(playlist_id):
        if os.path.isfile(file):
            continue
        filename = file.split('/')[-1]
        target_dir = '/'.join(file.split('/')[:-1])
        original_file = find(filename, music_dir)
        if not original_file:
            continue
        original_file_parent_dir = '/'.join(original_file.split('/')[:-1])
        files_to_move = [
            f for f in os.listdir(original_file_parent_dir)
            if os.path.isfile(original_file_parent_dir + '/' + f)
        ]
        if verbose:
            print('TO MOVE', original_file, target_dir, files_to_move)
        os.makedirs(target_dir, exist_ok=True)
        for move in files_to_move:
            os.rename(original_file_parent_dir + '/' + move, target_dir + '/' + move)
            if verbose:
                print('        MOVING', original_file_parent_dir + '/' + move, target_dir)
        os.rmdir(original_file_parent_dir)


def find_newer_than(base_path: str, seconds: int) -> List[Path]:
    return find_files(Path(base_path), lambda file: time() - os.path.getctime(file) < seconds)


def copy_newest_files(src_dir: str, target_dir: str, max_days: int, verbose: bool = False) -> None:
    to_copy = sorted(find_newer_than(src_dir, max_days * 24 * 60 * 60))
    for i, file in enumerate(to_copy):
        basedir = file.parent
        target_subdir = str(basedir).replace(src_dir, '').split(os.path.sep)
        target_path = Path(target_dir).joinpath(*target_subdir)
        target_path.mkdir(exist_ok=True)
        if verbose:
            print("{}/{} {}".format(i + 1, len(to_copy), str(file).replace(src_dir, '').strip('/')))
        if not file.joinpath(target_path).is_file():
            try:
                copy2(file, target_path)
            except OSError:
                pass


def parse_commandline(args: List[str]) -> Namespace:
    parser = ArgumentParser(
        description="Copy the first N existing files of an audacious playlist to a target folder"
    )
    subparsers = parser.add_subparsers(help='Available subcommands:', dest='subparser')
    # workaround for python issue https://bugs.python.org/issue26510
    subparsers.required = True  # type: ignore

    parser_copy = subparsers.add_parser(
        'copy', help='Copy files on the playlist to a specified target folder'
    )
    parser_copy.add_argument(
        '-p', '--playlist', type=str,
        help='ID of the playlist to copy (default: currently playing)'
    )
    parser_copy.add_argument(
        '-n', '--number', default=0, type=int,
        help='First N files to copy from the playlist (default: all)'
    )
    parser_copy.add_argument(
        '-r', '--renumber', action='store_true',
        help='Rename files to have playlist position prepended to file name'
    )
    parser_copy.add_argument('target', type=str, help='Name of the target folder')

    parser_restore = subparsers.add_parser(
        'restore', help='Move files back to the place in the file system they have on the playlist'
    )
    parser_restore.add_argument(
        '-p', '--playlist', type=str,
        help='ID of the playlist to copy (default: currently playing)'
    )

    parser_copy_newest = subparsers.add_parser(
        'copy_newest', help="Copy the latest files to a specified location"
    )
    parser_copy_newest.add_argument(
        '--max-age', type=int, help='Copy files newer than this many days'
    )
    parser_copy_newest.add_argument(
        '-s', '--source', type=str,
        default=os.path.expanduser('~/Music'),
        help='Source directory. Default: ' + os.path.expanduser('~/Music')
    )
    parser_copy_newest.add_argument('target', type=str, help='Name of the target folder')

    parser.add_argument(
        '-v', '--verbose', action='store_true'
    )

    return parser.parse_args(args)


def main() -> None:
    opts = parse_commandline(argv[1:])
    if opts.subparser == 'copy_newest':
        copy_newest_files(opts.source, opts.target, opts.max_age, verbose=opts.verbose)
    elif opts.subparser == 'restore':
        move_files_to_original_places(opts.playlist, verbose=opts.verbose)
    elif opts.subparser == 'copy':
        copy_playlist(opts.playlist, opts.number, opts.target, opts.verbose, opts.renumber)
    else:
        print("what are you doing?")


if __name__ == '__main__':
    main()



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import re
import sys
from math import ceil
from pathlib import Path
from time import sleep
from typing import Tuple, List, Dict, Optional

from pydub import AudioSegment
import requests
from tinytag import TinyTag

MIXCLOUD_MAX_FILESIZE = 512 * 1024 * 1024
MIXCLOUD_API_UPLOAD_URL = 'https://api.mixcloud.com/upload'
DEFAULT_CROSSFADE_MS = 1000
DEFAULT_MAX_RETRY = 100
DEFAULT_AUDIO_FILE_TYPES = ['flac', 'MP3', 'mp3', 'ogg', 'm4a']
# Mixcloud app:
# https://www.mixcloud.com/developers/Q29uc3VtZXI6MzY3Nw%253D%253D/
ACCESS_TOKEN_FILE = '.mixcloud_access_token'
MP3_KBIT_RATE = 128


def bytes_per_second(mp3_kbit_rate: int = MP3_KBIT_RATE) -> int:
    return mp3_kbit_rate // 8 * 1024 * 2


class Mix:  # pylint: disable=too-many-instance-attributes

    @classmethod
    def create(  # pylint: disable=too-many-arguments
            cls, basedir: Path, patterns: Tuple[str, ...],
            verbose: bool = False, strict: bool = False,
            crossfade_ms: int = DEFAULT_CROSSFADE_MS, title: str = None
    ) -> "Mix":
        files, length = cls.scan(basedir, patterns)
        if bytes_per_second() * length < MIXCLOUD_MAX_FILESIZE:
            return Mix(basedir, patterns, verbose, strict, crossfade_ms, title)
        return MultiMix(
            basedir, files, length, verbose, strict, crossfade_ms, title
        )

    @classmethod
    def scan(cls, basedir: Path, patterns: Tuple[str, ...]) -> Tuple[List[str], int]:
        audio_files = sorted([f for p in patterns for f in basedir.glob(p)])
        length = sum([TinyTag.get(str(audio_file)).duration for audio_file in audio_files])
        return [f.name for f in audio_files], int(length)

    @staticmethod
    def _read_access_token() -> str:
        with open(ACCESS_TOKEN_FILE, 'r') as file:
            return file.read().strip()

    def __init__(  # pylint: disable=too-many-arguments
            self, basedir: Path, patterns: Tuple[str, ...],
            verbose: bool = False, strict: bool = False,
            crossfade_ms: int = DEFAULT_CROSSFADE_MS, title: str = None
    ) -> None:
        self._basedir = basedir
        self._verbose = verbose
        self._strict = strict
        self._crossfade_ms = crossfade_ms
        self._title = title
        self._incomplete = False
        self._track_info: List[Dict] = []
        self._access_token = self._read_access_token()
        audio_files = sorted([f for p in patterns for f in self._basedir.glob(p)])
        self._audio = self._import_audio(audio_files)

    @property
    def valid(self):
        return not self._incomplete

    @property
    def tags(self) -> List[str]:
        if (self._basedir / 'tags.txt').exists():
            with (self._basedir / 'tags.txt').open() as file:
                return list(line for line in (line.strip() for line in file) if line)
        if self._strict:
            raise ValueError('No tags found')
        print('WARNING - no tags found')
        self._incomplete = True
        return ['Testing Mixcloud API']

    @property
    def title(self) -> str:
        if self._title:
            return self._title
        title = re.sub(r'^\d+ - ', '', self._basedir.resolve().name)
        return f"Test - don't bother playing ({title})" if self._incomplete else title

    @property
    def description(self) -> str:
        if (self._basedir / 'description.txt').exists():
            with (self._basedir / 'description.txt').open() as file:
                return file.read().strip()
        if self._strict:
            raise ValueError('No description found')
        print('WARNING - no description found')
        self._incomplete = True
        return 'Test test test'

    @property
    def picture(self) -> Optional[Path]:
        """Currently just returns the first JPEG or PNG. Room for improvement!"""
        try:
            return next(self._basedir.glob('*.*p*g'))
        except StopIteration as error:
            if self._strict:
                raise ValueError(f"No picture in {self._basedir}") from error
            print('WARNING - no picture found')
            self._incomplete = True
            return None

    def _import_audio(self, audio_files: List[Path]) -> AudioSegment:
        audio = AudioSegment.empty()
        for audio_file in audio_files:
            self._track_info.append(self._get_track_info(audio_file))
            if self._verbose:
                print(audio_file.name)
            track = AudioSegment.from_file(audio_file)
            track = track.normalize()
            audio = audio.append(
                track, crossfade=self._crossfade_ms if len(audio) > self._crossfade_ms else 0
            )

        return audio

    def _get_track_info(self, audio_file: Path) -> Dict:
        tags = TinyTag.get(str(audio_file))
        if tags.artist is None or tags.title is None:
            if self._strict:
                raise ValueError(f"Incomplete tags for {audio_file}")
            self._incomplete = True
            return {'artist': '???', 'title': '???', 'length': tags.duration}
        return {
            'artist': tags.artist, 'title': tags.title, 'length': tags.duration,
            'filename': audio_file.name
        }

    def export(self, name: Path = Path('mix.mp3')) -> None:
        mix_file = self._basedir / name
        audio_format = name.suffix[1:]
        if self._verbose:
            print(f'Exporting to {mix_file} with bitrate {MP3_KBIT_RATE} kbps')
        self._audio.export(
            mix_file, format=audio_format, parameters=["-q:a", "0"], bitrate=f'{MP3_KBIT_RATE}k'
        )

    def upload(self, name: Path = Path('mix.mp3'), max_retry: int = DEFAULT_MAX_RETRY) -> None:
        url = MIXCLOUD_API_UPLOAD_URL + '/?access_token=' + self._access_token
        mix_file = self._basedir / name
        files = {
            'mp3': ('mix.mp3', mix_file.open('rb'), 'audio/mpeg'),
        }
        if self.picture:
            picture_type = self.picture.suffix
            files['picture'] = (
                'picture' + picture_type, self.picture.open('rb'),
                f'image/{"png" if picture_type == ".png" else "jpeg"}'
            )

        data = {
            'name': self.title,
            'description': self.description,
            'percentage_music': 100
        }
        self._add_tags(data)
        self._add_track_info(data)
        if self._verbose:
            size = mix_file.stat().st_size + self.picture.stat().st_size
            print(
                f'Uploading {size // 1024:,d} kBytes '
                f'({len(self._audio) // 60000}:{len(self._audio) % 60000 // 1000:02} minutes) '
                f'as {self.title}'
            )

        try:
            response = requests.post(url, files=files, data=data)
        except requests.exceptions.ConnectionError as error:
            if self._verbose:
                print(self.connection_error_message(error, max_retry))
            if max_retry > 0:
                self.upload(name, max_retry=max_retry - 1)
                return
            sys.exit(1)

        if self._verbose:
            print(self.response_message(response))
        if self.is_rate_limited(response) and max_retry > 0:
            sleep(int(response.json()['error']['retry_after']))
            self.upload(name, max_retry=max_retry - 1)
        else:
            sys.exit(1)

    @staticmethod
    def is_rate_limited(response: requests.Response) -> bool:
        return response.json().get('error', {}).get('type') == 'RateLimitException'

    @staticmethod
    def response_message(response: requests.Response) -> str:
        if response.status_code == 200:
            return f"{response.status_code}: {response.json()['result']['message']}"
        return f"{response.status_code}: {response.json()['error']}"

    @staticmethod
    def connection_error_message(error: requests.exceptions.ConnectionError, max_retry: int) -> str:
        message = f"{error.args[0].args[1].args[1]}: {error.args[0].args[0]} "
        if max_retry > 0:
            message += f"Retrying {max_retry} time{'s' if max_retry > 1 else ''}."
        else:
            message += "Giving up."
        return message

    def _add_tags(self, data: Dict) -> None:
        for i, tag in enumerate(self.tags):
            data[f"tags-{i}-tag"] = tag

    def _add_track_info(self, data: Dict) -> None:
        start_time = 0
        for i, track_info in enumerate(self._track_info):
            data[f"sections-{i}-artist"] = track_info['artist']
            data[f"sections-{i}-song"] = track_info['title']
            data[f"sections-{i}-start_time"] = int(start_time)
            start_time += (track_info['length'] - self._crossfade_ms / 1000)


# noinspection PyMissingConstructor
class MultiMix(Mix):

    def __init__(  # pylint: disable=super-init-not-called, too-many-arguments
            self, basedir: Path, files: List[str], total_length: int,
            verbose: bool = False, strict: bool = False,
            crossfade_ms: int = DEFAULT_CROSSFADE_MS, title: str = None
    ) -> None:
        self._basedir = basedir
        self._title = title
        self._incomplete = False
        self._mix_parts: List[Mix] = []
        self._part_paths: List[Path] = []
        oversize_factor = ceil(total_length * bytes_per_second() / MIXCLOUD_MAX_FILESIZE)
        chunk_size = len(files) // oversize_factor
        for i in range(oversize_factor):
            part_files = tuple(files[i * chunk_size:(i + 1) * chunk_size])
            part_name = f"{self.title} Part {i + 1}"
            if verbose:
                print(part_name)
            mix_part = Mix(
                basedir, part_files, verbose=verbose, strict=strict, crossfade_ms=crossfade_ms,
                title=part_name
            )
            self._mix_parts.append(mix_part)

    @property
    def parts(self):
        return self._mix_parts

    def upload(self, name: Path = Path('mix.mp3'), max_retry: int = DEFAULT_MAX_RETRY) -> None:
        for mix_part, part_path in zip(self._mix_parts, self._part_paths):
            mix_part.upload(part_path, max_retry)

    def export(self, name: Path = Path('mix.mp3')) -> None:
        for i, mix_part in enumerate(self._mix_parts):
            part_path = Path(f"{name.stem}_{i + 1}{name.suffix}")
            mix_part.export(part_path)
            self._part_paths.append(part_path)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import os
from subprocess import check_output
from typing import List, Optional
from urllib.parse import unquote

DEFAULT_AUDACIOUS_CONFIG_DIR = os.path.join(
    os.environ['HOME'], '.config', 'audacious'
)


class AudaciousTools:
    """Tools for working with the audacious media player"""

    PLAYLIST_DIR_NAME = 'playlists'
    PLAYLIST_EXTENSION = '.audpl'
    FILE_LINE_PREFIX = 'uri=file://'

    def __init__(
            self, base_config_dir: str = DEFAULT_AUDACIOUS_CONFIG_DIR, file_base_dir: str = '.'
    ) -> None:
        """
        :param base_config_dir: Directory containing the audacious configuration
        """
        self._base_directory = base_config_dir
        self._playlist_dir = find_first_dir(self.PLAYLIST_DIR_NAME, self._base_directory)
        self._file_base_dir = file_base_dir

    def get_currently_playing_playlist_id(self) -> str:
        order = self.get_playlist_order()
        return order[AudaciousTools._currently_playing_playlist_number() - 1]

    def get_playlist_order(self) -> List[str]:
        """All playlist ids for this audacious instance, sorted in tab order"""
        with open(self._playlist_order_file_path()) as order_file:
            return order_file.readlines()[0].split(' ')

    @property
    def playlist_directory(self) -> str:
        """Directory where the playlists are for this audacious instance"""
        return self._playlist_dir

    def files_in_playlist(self, playlist_id: str, existing_only: bool = False) -> List[str]:
        """
        :param playlist_id: Playlist ID (filename)
        :param existing_only: only return files that exist in filesystem
        :return: All actually existing files in that playlist
        """
        lines = self._read_playlist(playlist_id)
        files = self._file_entries(lines)
        return existing_files(files) if existing_only else files

    def get_files_to_copy(self, number: int, playlist_id: str) -> List[str]:
        files_to_copy = self.files_in_playlist(playlist_id)
        return files_to_copy[:number] if number else files_to_copy

    def _playlist_order_file_path(self) -> str:
        return os.path.join(self.playlist_directory, 'order')

    def _read_playlist(self, playlist_id: str) -> List[str]:
        with open(self._playlist_file_path(playlist_id)) as playlist_file:
            return playlist_file.readlines()

    def _playlist_file_path(self, playlist_id: str) -> str:
        return os.path.join(
            self.playlist_directory, playlist_id + AudaciousTools.PLAYLIST_EXTENSION
        )

    def _file_entries(self, lines: List[str]) -> List[str]:
        return [
            os.path.join(
                self._file_base_dir, unquote(line[len(AudaciousTools.FILE_LINE_PREFIX):]).strip()
            )
            for line in lines if line.startswith(AudaciousTools.FILE_LINE_PREFIX)
        ]

    @staticmethod
    def _currently_playing_playlist_number() -> int:
        return int(check_output(['audtool', 'current-playlist']))


def find_first_dir(name: str, path: str) -> Optional[str]:
    for root, dirs, _ in os.walk(path):
        if name in dirs:
            return os.path.join(root, name)
    return None


def existing_files(files: List[str]) -> List[str]:
    return [file for file in files if os.path.isfile(file)]



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import os
from pathlib import Path
from typing import Any, Callable, List

Condition = Callable[[Path], bool]


def find_potential_files(base_path: Path, condition: Condition) -> List[Path]:
    return [
        Path(root).joinpath(file)
        for root, _, files in os.walk(str(base_path))
        for file in files
        if condition(Path(root).joinpath(file))
    ]


def find_files(base_path: Path, condition: Condition) -> List[Path]:
    return [file for file in find_potential_files(base_path, condition) if file.exists()]


def find_dirs(base_path: Path, condition: Condition) -> List[Path]:
    return [
        Path(root).joinpath(dir)
        for root, dirs, _ in os.walk(str(base_path))
        for dir in dirs
        if Path(root).joinpath(dir).exists()
        if condition(Path(root).joinpath(dir))
    ]


def print_utf8_safe(*args: Any) -> None:
    try:
        print(*args)
    except UnicodeEncodeError as error:
        raise ValueError(args) from error



import re
from pathlib import Path

from .base_filename_cleaner import BaseFilenameCleaner
from .command import Command, Commands, Move
from .util import find_files


class JunkRemover(BaseFilenameCleaner):

    JUNK_TO_REMOVE = {
        ' $': '',     # space(s) at beginning and before ".mp3"
        '-$': '',     # "-.mp3", " - .mp3"
        '^-': '',     # dash at beginning
        '^ ': '',     # space at beginning
        r'^\.': '',   # dot at beginning
        '--': '-',    # --
        '- -': '-',   # stray double dashes
        '_': ' ',     # underscores
        '-,': '-',    # dash before comma (BECAUSE THAT HAPPENS AND IT FUCKS EVERYTHING UP JFC)
        r' -(\S)': r' - \1',  # immediately leading dash with space before
        r'(\S)- ': r'\1 - ',  # immediately trailing dash with space after
        '  ': ' ',    # double spaces
        ' ,': ',',    # space before comma
        r'\[\]': '',
        r'\(\)': '',
    }

    def fix_commands(self) -> Commands:
        def has_junk(filename: Path) -> bool:
            return self.is_music_file(filename) and \
                any(re.search(s, self.filename_base(filename)) for s in self.JUNK_TO_REMOVE)

        mismatches = sorted(find_files(self._base_directory, has_junk))
        return [self.fix_command_for_file(mismatch) for mismatch in mismatches]

    def fix_command_for_file(self, mismatch: Path) -> Command:
        root = mismatch.parent
        fixed = self.filename_base(mismatch)
        extension = mismatch.suffix
        matches = []
        changed = True
        while changed:
            changed = False
            for search, replace in self.JUNK_TO_REMOVE.items():
                new_fixed = re.sub(search, replace, fixed)
                if new_fixed != fixed:
                    changed = True
                    matches.append(search)
                fixed = new_fixed
        return Move(mismatch, root.joinpath(fixed + extension), matches)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import sys
from argparse import ArgumentParser
from calendar import timegm
from datetime import date, datetime
from glob import glob
from pickle import dump, load
from os import environ
from time import sleep
from typing import Dict, Optional

from pylast import LastFMNetwork, User, WSError

MAX_RETRY_WAIT = 600

"""
export PYLAST_USERNAME=ENTER_YOURS_HERE
export PYLAST_API_KEY=ENTER_YOURS_HERE
export PYLAST_API_SECRET=ENTER_YOURS_HERE
"""


def from_datetime(convert_datetime: datetime) -> int:
    return timegm(convert_datetime.utctimetuple())


def year_range(year: int) -> Dict[str, int]:
    start = datetime(year, 1, 1, 0, 0, 0)
    end = datetime(year, 12, 31, 23, 59, 59, 999999)
    return dict(time_from=from_datetime(start), time_to=from_datetime(end))


def start_year(user: User) -> int:
    return date.fromtimestamp(user.get_unixtime_registered()).year


def get_data(
        user: User, year: Optional[int], limit: Optional[int], retry_wait_time: int = 1
) -> Dict:
    year_dates = year_range(year)
    try:
        return dict(
            time=datetime.now(),
            play_count=user.get_playcount(),
            tracks=user.get_recent_tracks(limit=limit, **year_dates),
            loved=user.get_loved_tracks(limit=limit),
            top_tracks=user.get_top_tracks(limit=limit),
            top_artists=user.get_top_artists(limit=limit),
            top_albums=user.get_top_albums(limit=limit),
            top_tags=user.get_top_tags(limit=limit),
        )
    except WSError as error:
        # pylast has a feature to enable rate limiting but it doesn't seem to work reliably
        if "Rate Limit Exceeded" in str(error) and retry_wait_time < MAX_RETRY_WAIT:
            print(f"{error}, retrying in {retry_wait_time}s")
            sleep(retry_wait_time)
            retry_wait_time *= 2
            return get_data(user, year, limit, 2 * retry_wait_time)
        raise


def combine_data() -> None:
    separate_user_data = {}
    user_data = {}
    for filename in glob('backup_lastfm_20??.pickle'):
        with open(filename, 'rb') as file:
            separate_user_data[filename] = load(file)
        user_data = separate_user_data[filename]
    user_data['tracks'] = sum([data['tracks'] for data in separate_user_data.values()], [])
    with open('backup_lastfm_all.pickle', 'wb') as backup_file:
        dump(user_data, backup_file)


def main() -> None:
    parser = ArgumentParser(
        description="Creates a pickle file from a last.fm user's data"
    )
    parser.add_argument(
        '-l', '--limit', type=int, default=None,
        help='Max. number of entries to request'
    )
    parser.add_argument(
        '-y', '--year', type=int, default=None,
        help='Year the data is requested for'
    )
    parser.add_argument(
        '-c', '--combine', action='store_true',
        help='Combine data sets present in current folder'
    )

    args = parser.parse_args()
    if args.combine:
        combine_data()
        sys.exit(0)

    network = LastFMNetwork(
        api_key=environ['PYLAST_API_KEY'], api_secret=environ['PYLAST_API_SECRET']
    )

    user = network.get_user(environ['PYLAST_USERNAME'])
    years = [args.year] if args.year else range(start_year(user), datetime.now().year)

    for year in years:
        user_data = get_data(user, year, args.limit)
        with open(f'backup_lastfm_{year}.pickle', 'wb') as backup_file:
            dump(user_data, backup_file)


if __name__ == '__main__':
    main()



import re
from pathlib import Path

from .base_filename_cleaner import BaseFilenameCleaner
from .command import Command, Commands, Move, Nothing
from .util import find_files, print_utf8_safe


class NumberingFixer(BaseFilenameCleaner):

    PATTERNS_TO_FIX = [
        r'(\d\d)\2\s+-\s+([^/]+)',       # 0101 - blah (note: regex is prefixed by group for path)
        r'\s*(\d{1,3}-\d)\s+([^-][^/]+)',  # 01-1 blah
        r'\s*(\d-\d{1,2})\s+([^-][^/]+)',  # 1-01 blah
        r'\s*(\d\d-\d\d)\s+([^-][^/]+)',   # 01-01 blah
        r'\s*(\d{1,4})\s+([^/]+)',       # 01 blah
        r'\s*(\d{1,3})\.\s+([^/]+)',     # 01. blah
        r'\s*(\d{1,3})\.([^/]+)',        # 01.blah
        r'\s*(\d{1,3})--([^/]+)',        # 01--blah
        r'\s*(\d{1,3})-\s*(\D[^/]+)',    # 01-blah or 01- blah
        r'\s*-(\d{1,3})-([^/]+)',        # -01-blah
        r'\s*-\s*(\d{1,3})\.\s*([^/]+)',  # - 01. blah
        r'\s*(\d{1,3})_([^/]+)',         # 01_blah
        r'\s*\[(\d{1,3})\]([^/]+)',      # [01]blah
        r'\s*(\d{1,3})\]([^/]+)',        # 01]blah
        r'\s*\((\d{1,3})\)\s*([^/]+)',   # (01)blah
        r'\s*(\d{1,4})(\D[^/]*)',        # 01blah
        r'\s*([a-z]\d{1,2})\s+([^/]+)',  # a1 blah
        r'\s*([a-z]\d)-([^/]+)',         # a1-blah
        r'\s*([a-z]\d)\.([^/]+)',        # a1.blah
        r'\s*\[([a-z]\d)\]([^/]+)',      # [a1]blah
        r'\s*([a-z]\d)\]([^/]+)',        # a1]blah
        r'\s*\(([a-z]\d)\)([^/]+)',      # (a1)blah
        r'\s*([a-z]\d{1,2})(\D[^/]+)',   # a1blah
    ]

    def fix_commands(self) -> Commands:
        def has_screwy_numbering(filename: Path) -> bool:
            base = self.filename_base(filename)
            return self.is_music_file(filename) and bool(
                re.search(r'\d+', base) and
                not re.search(r'^\d{1,4} - [^/]+', base, flags=re.IGNORECASE) and
                not re.search(r'^\d{1,4}$', base, flags=re.IGNORECASE) and
                not re.search(r'^\d{1,2}-\d{1,2}$', base, flags=re.IGNORECASE) and
                not re.search(r'^\d{1,2}-\d{1,2} - [^/]+', base, flags=re.IGNORECASE) and
                not re.search(r'^[a-z]\d{1,2} - [^/]+', base, flags=re.IGNORECASE) or
                re.search(r'^(\d\d)\1 - [^/]+', base, flags=re.IGNORECASE)
            )

        mismatches = sorted(find_files(self._base_directory, has_screwy_numbering))
        return [self._fix_numbering_for_file(file) for file in mismatches]

    def _fix_numbering_for_file(self, file: Path) -> Command:
        for extension in self.MUSIC_EXTENSIONS:
            if re.match(r'(.*)/(\d{1,4})\.' + extension, str(file), flags=re.IGNORECASE):
                return Nothing()
            for pattern in self.PATTERNS_TO_FIX:
                match = re.search(
                    '(.*)/' + pattern + r'\.' + extension, str(file), flags=re.IGNORECASE
                )
                if match:
                    return Move(
                        Path(file),
                        Path(f"{match.group(1)}/{match.group(2)} - {match.group(3)}.{extension}"),
                        [pattern]
                    )
        if self._verbose:
            print_utf8_safe('-' * 8, file)
        return Nothing()



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import List

from media_tools.util.mixcloud import (
    Mix, DEFAULT_CROSSFADE_MS, DEFAULT_MAX_RETRY, DEFAULT_AUDIO_FILE_TYPES
)


def parse_commandline(args: List[str]) -> Namespace:
    parser = ArgumentParser(
        description="Creates a mix from audio files and uploads it to Mixcloud"
    )
    parser.add_argument(
        '-d', '--directory', type=str, required=True, help='Directory containing the mix'
    )
    parser.add_argument(
        '-e', '--extensions', nargs='+', default=DEFAULT_AUDIO_FILE_TYPES,
        help='List of extensions considered for the mix'
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true'
    )
    parser.add_argument(
        '-s', '--strict', action='store_true', help='Fail if any required data are missing'
    )
    parser.add_argument(
        '-c', '--crossfade-ms', type=int, default=DEFAULT_CROSSFADE_MS,
        help='Milliseconds overlap between tracks'
    )
    parser.add_argument(
        '-r', '--max-retry', type=int, default=DEFAULT_MAX_RETRY,
        help='Maximum number of retries for failing uploads'
    )
    return parser.parse_args(args)


def main() -> None:
    args: List[str] = sys.argv[1:]
    opts = parse_commandline(args)
    mix = Mix.create(
        Path(opts.directory), tuple(f'?? - *.{ext}' for ext in opts.extensions),
        crossfade_ms=opts.crossfade_ms, verbose=opts.verbose, strict=opts.strict
    )
    mix.export()
    mix.upload(max_retry=opts.max_retry)


if __name__ == '__main__':
    main()
